@@ -1,7 +1,7 @@
 /// @file htslib/hts.h
 /// Format-neutral I/O, indexing, and iterator API functions.
 /*
-    Copyright (C) 2012-2016 Genome Research Ltd.
+    Copyright (C) 2012-2019 Genome Research Ltd.
     Copyright (C) 2010, 2012 Broad Institute.
     Portions copyright (C) 2003-2006, 2008-2010 by Heng Li <lh3@live.co.uk>
 
@@ -30,6 +30,7 @@ DEALINGS IN THE SOFTWARE.  */
 
 #include <stddef.h>
 #include <stdint.h>
+#include <inttypes.h>
 
 #include "hts_defs.h"
 #include "hts_log.h"
@@ -38,6 +39,15 @@ DEALINGS IN THE SOFTWARE.  */
 extern "C" {
 #endif
 
+// Separator used to split HTS_PATH (for plugins); REF_PATH (cram references)
+#if defined(_WIN32) || defined(__MSYS__)
+#define HTS_PATH_SEPARATOR_CHAR ';'
+#define HTS_PATH_SEPARATOR_STR  ";"
+#else
+#define HTS_PATH_SEPARATOR_CHAR ':'
+#define HTS_PATH_SEPARATOR_STR  ":"
+#endif
+
 #ifndef HTS_BGZF_TYPEDEF
 typedef struct BGZF BGZF;
 #define HTS_BGZF_TYPEDEF
@@ -45,10 +55,11 @@ typedef struct BGZF BGZF;
 struct cram_fd;
 struct hFILE;
 struct hts_tpool;
+struct sam_hdr_t;
 
 #ifndef KSTRING_T
 #define KSTRING_T kstring_t
-typedef struct __kstring_t {
+typedef struct kstring_t {
     size_t l, m;
     char *s;
 } kstring_t;
@@ -60,7 +71,7 @@ typedef struct __kstring_t {
 
 /**
  * @hideinitializer
- * Macro to expand a dynamic array of a given type
+ * Deprecated macro to expand a dynamic array of a given type
  *
  * @param         type_t The type of the array elements
  * @param[in]     n      Requested number of elements of type type_t
@@ -68,9 +79,12 @@ typedef struct __kstring_t {
  * @param[in,out] ptr    Pointer to the array
  *
  * @discussion
+ * Do not use this macro.  Use hts_resize() instead as allows allocation
+ * failures to be handled more gracefully.
+ *
  * The array *ptr will be expanded if necessary so that it can hold @p n
  * or more elements.  If the array is expanded then the new size will be
- * written to @p m and the value in @ptr may change.
+ * written to @p m and the value in @p ptr may change.
  *
  * It must be possible to take the address of @p ptr and @p m must be usable
  * as an lvalue.
@@ -99,6 +113,9 @@ typedef struct __kstring_t {
  * @param[in,out] ptr    Pointer to the array
  *
  * @discussion
+ * Do not use this macro.  Use hts_resize() instead as allows allocation
+ * failures to be handled more gracefully.
+ *
  * As for hts_expand(), except the bytes that make up the array elements
  * between the old and new values of @p m are set to zero using memset().
  *
@@ -118,6 +135,51 @@ typedef struct __kstring_t {
         }                                                               \
     } while (0)
 
+// For internal use (by hts_resize()) only
+HTSLIB_EXPORT
+int hts_resize_array_(size_t, size_t, size_t, void *, void **, int,
+                      const char *);
+
+#define HTS_RESIZE_CLEAR 1
+
+/**
+ * @hideinitializer
+ * Macro to expand a dynamic array of a given type
+ *
+ * @param         type_t    The type of the array elements
+ * @param[in]     num       Requested number of elements of type type_t
+ * @param[in,out] size_ptr  Pointer to where the size (in elements) of the
+                            array is stored.
+ * @param[in,out] ptr       Location of the pointer to the array
+ * @param[in]     flags     Option flags
+ *
+ * @return        0 for success, or negative if an error occurred.
+ *
+ * @discussion
+ * The array *ptr will be expanded if necessary so that it can hold @p num
+ * or more elements.  If the array is expanded then the new size will be
+ * written to @p *size_ptr and the value in @p *ptr may change.
+ *
+ * If ( @p flags & HTS_RESIZE_CLEAR ) is set, any newly allocated memory will
+ * be cleared.
+ */
+
+#define hts_resize(type_t, num, size_ptr, ptr, flags)       \
+    ((num) > (*(size_ptr))                                  \
+     ? hts_resize_array_(sizeof(type_t), (num),             \
+                         sizeof(*(size_ptr)), (size_ptr),   \
+                         (void **)(ptr), (flags), __func__) \
+     : 0)
+
+/**
+ * Wrapper function for free(). Enables memory deallocation across DLL
+ * boundary. Should be used by all applications, which are compiled
+ * with a different standard library than htslib and call htslib
+ * methods that return dynamically allocated data.
+ */
+HTSLIB_EXPORT
+void hts_free(void *ptr);
+
 /************
  * File I/O *
  ************/
@@ -140,11 +202,13 @@ enum htsExactFormat {
     sam, bam, bai, cram, crai, vcf, bcf, csi, gzi, tbi, bed,
     htsget,
     json HTS_DEPRECATED_ENUM("Use htsExactFormat 'htsget' instead") = htsget,
+    empty_format,  // File is empty (or empty after decompression)
+    fasta_format, fastq_format, fai_format, fqi_format,
     format_maximum = 32767
 };
 
 enum htsCompression {
-    no_compression, gzip, bgzf, custom,
+    no_compression, gzip, bgzf, custom, bzip2_compression,
     compression_maximum = 32767
 };
 
@@ -157,6 +221,9 @@ typedef struct htsFormat {
     void *specific;  // format specific options; see struct hts_opt.
 } htsFormat;
 
+struct __hts_idx_t;
+typedef struct __hts_idx_t hts_idx_t;
+
 // Maintainers note htsFile cannot be an opaque structure because some of its
 // fields are part of libhts.so's ABI (hence these fields must not be moved):
 //  - fp is used in the public sam_itr_next()/etc macros
@@ -174,7 +241,11 @@ typedef struct {
         struct cram_fd *cram;
         struct hFILE *hfile;
     } fp;
+    void *state;  // format specific state information
     htsFormat format;
+    hts_idx_t *idx;
+    const char *fnidx;
+    struct sam_hdr_t *bam_header;
 } htsFile;
 
 // A combined thread pool and queue allocation size.
@@ -256,6 +327,12 @@ typedef struct hts_opt {
 
 #define HTS_FILE_OPTS_INIT {{0},0}
 
+/*
+ * Explicit index file name delimiter, see below
+ */
+#define HTS_IDX_DELIM "##idx##"
+
+
 /**********************
  * Exported functions *
  **********************/
@@ -266,6 +343,7 @@ typedef struct hts_opt {
  * Returns 0 on success;
  *        -1 on failure.
  */
+HTSLIB_EXPORT
 int hts_opt_add(hts_opt **opts, const char *c_arg);
 
 /*
@@ -274,11 +352,13 @@ int hts_opt_add(hts_opt **opts, const char *c_arg);
  * Returns 0 on success
  *        -1 on failure
  */
+HTSLIB_EXPORT
 int hts_opt_apply(htsFile *fp, hts_opt *opts);
 
 /*
  * Frees an hts_opt list.
  */
+HTSLIB_EXPORT
 void hts_opt_free(hts_opt *opts);
 
 /*
@@ -289,6 +369,7 @@ void hts_opt_free(hts_opt *opts);
  * Returns 0 on success
  *        -1 on failure.
  */
+HTSLIB_EXPORT
 int hts_parse_format(htsFormat *opt, const char *str);
 
 /*
@@ -302,6 +383,7 @@ int hts_parse_format(htsFormat *opt, const char *str);
  * Returns 0 on success
  *        -1 on failure.
  */
+HTSLIB_EXPORT
 int hts_parse_opt_list(htsFormat *opt, const char *str);
 
 /*! @abstract Table for converting a nucleotide character to 4-bit encoding.
@@ -326,14 +408,28 @@ extern const int seq_nt16_int[];
   @return    For released versions, a string like "N.N[.N]"; or git describe
   output if using a library built within a Git repository.
 */
+HTSLIB_EXPORT
 const char *hts_version(void);
 
+/*!
+  @abstract  Compile-time HTSlib version number, for use in #if checks
+  @return    For released versions X.Y[.Z], an integer of the form XYYYZZ;
+  useful for preprocessor conditionals such as
+      #if HTS_VERSION >= 101000  // Check for v1.10 or later
+*/
+// Maintainers: Bump this in the final stage of preparing a new release.
+// Immediately after release, bump ZZ to 90 to distinguish in-development
+// Git repository builds from the release; you may wish to increment this
+// further when significant features are merged.
+#define HTS_VERSION 101000
+
 /*!
   @abstract    Determine format by peeking at the start of a file
   @param fp    File opened for reading, positioned at the beginning
   @param fmt   Format structure that will be filled out on return
   @return      0 for success, or negative if an error occurred.
 */
+HTSLIB_EXPORT
 int hts_detect_format(struct hFILE *fp, htsFormat *fmt);
 
 /*!
@@ -341,11 +437,15 @@ int hts_detect_format(struct hFILE *fp, htsFormat *fmt);
   @param fmt   Format structure holding type, version, compression, etc.
   @return      Description string, to be freed by the caller after use.
 */
+HTSLIB_EXPORT
 char *hts_format_description(const htsFormat *format);
 
 /*!
-  @abstract       Open a SAM/BAM/CRAM/VCF/BCF/etc file
-  @param fn       The file name or "-" for stdin/stdout
+  @abstract       Open a sequence data (SAM/BAM/CRAM) or variant data (VCF/BCF)
+                  or possibly-compressed textual line-orientated file
+  @param fn       The file name or "-" for stdin/stdout. For indexed files
+                  with a non-standard naming, the file name can include the
+                  name of the index file delimited with HTS_IDX_DELIM
   @param mode     Mode matching / [rwa][bceguxz0-9]* /
   @discussion
       With 'r' opens for reading; any further format mode letters are ignored
@@ -370,6 +470,7 @@ char *hts_format_description(const htsFormat *format);
       [rw]z  .. compressed VCF
       [rw]   .. uncompressed VCF
 */
+HTSLIB_EXPORT
 htsFile *hts_open(const char *fn, const char *mode);
 
 /*!
@@ -386,6 +487,7 @@ htsFile *hts_open(const char *fn, const char *mode);
       like pointers to the reference or information on compression levels,
       block sizes, etc.
 */
+HTSLIB_EXPORT
 htsFile *hts_open_format(const char *fn, const char *mode, const htsFormat *fmt);
 
 /*!
@@ -393,6 +495,7 @@ htsFile *hts_open_format(const char *fn, const char *mode, const htsFormat *fmt)
   @param fn       The already-open file handle
   @param mode     Open mode, as per hts_open()
 */
+HTSLIB_EXPORT
 htsFile *hts_hopen(struct hFILE *fp, const char *fn, const char *mode);
 
 /*!
@@ -400,6 +503,7 @@ htsFile *hts_hopen(struct hFILE *fp, const char *fn, const char *mode);
   @param fp  The file handle to be closed
   @return    0 for success, or negative if an error occurred.
 */
+HTSLIB_EXPORT
 int hts_close(htsFile *fp);
 
 /*!
@@ -407,6 +511,7 @@ int hts_close(htsFile *fp);
   @param fp  The file handle
   @return    Read-only pointer to the file's htsFormat.
 */
+HTSLIB_EXPORT
 const htsFormat *hts_get_format(htsFile *fp);
 
 /*!
@@ -414,6 +519,7 @@ const htsFormat *hts_get_format(htsFile *fp);
   @ param format  Format structure containing the file type.
   @ return        A string ("sam", "bam", etc) or "?" for unknown formats.
  */
+HTSLIB_EXPORT
 const char *hts_format_file_extension(const htsFormat *format);
 
 /*!
@@ -423,9 +529,12 @@ const char *hts_format_file_extension(const htsFormat *format);
   @param ... Optional arguments, dependent on the option used.
   @return    0 for success, or negative if an error occurred.
 */
+HTSLIB_EXPORT
 int hts_set_opt(htsFile *fp, enum hts_fmt_option opt, ...);
 
+HTSLIB_EXPORT
 int hts_getline(htsFile *fp, int delimiter, kstring_t *str);
+HTSLIB_EXPORT
 char **hts_readlines(const char *fn, int *_n);
 /*!
     @abstract       Parse comma-separated list or read list from a file
@@ -435,6 +544,7 @@ char **hts_readlines(const char *fn, int *_n);
     @return         NULL on failure or pointer to newly allocated array of
                     strings
 */
+HTSLIB_EXPORT
 char **hts_readlist(const char *fn, int is_file, int *_n);
 
 /*!
@@ -445,6 +555,7 @@ char **hts_readlist(const char *fn, int is_file, int *_n);
   @notes     This function creates non-shared threads for use solely by fp.
              The hts_set_thread_pool function is the recommended alternative.
 */
+HTSLIB_EXPORT
 int hts_set_threads(htsFile *fp, int n);
 
 /*!
@@ -453,6 +564,7 @@ int hts_set_threads(htsFile *fp, int n);
   @param p   A pool of worker threads, previously allocated by hts_create_threads().
   @return    0 for success, or negative if an error occurred.
 */
+HTSLIB_EXPORT
 int hts_set_thread_pool(htsFile *fp, htsThreadPool *p);
 
 /*!
@@ -461,6 +573,7 @@ int hts_set_thread_pool(htsFile *fp, htsThreadPool *p);
   @param fp  The file handle
   @param n   The size of cache, in bytes
 */
+HTSLIB_EXPORT
 void hts_set_cache_size(htsFile *fp, int n);
 
 /*!
@@ -470,6 +583,7 @@ void hts_set_cache_size(htsFile *fp, int n);
       Called before *_hdr_read(), this provides the name of a .fai file
       used to provide a reference list if the htsFile contains no @SQ headers.
 */
+HTSLIB_EXPORT
 int hts_set_fai_filename(htsFile *fp, const char *fn_aux);
 
 
@@ -483,6 +597,7 @@ int hts_set_fai_filename(htsFile *fp, const char *fn_aux);
   @discussion
       Check if the BGZF end-of-file (EOF) marker is present
 */
+HTSLIB_EXPORT
 int hts_check_EOF(htsFile *fp);
 
 /************
@@ -509,12 +624,26 @@ When REST or NONE is used, idx is also ignored and may be NULL.
 #define HTS_FMT_TBI 2
 #define HTS_FMT_CRAI 3
 
-struct __hts_idx_t;
-typedef struct __hts_idx_t hts_idx_t;
+// Almost INT64_MAX, but when cast into a 32-bit int it's
+// also INT_MAX instead of -1.  This avoids bugs with old code
+// using the new hts_pos_t data type.
+#define HTS_POS_MAX ((((int64_t)INT_MAX)<<32)|INT_MAX)
+#define HTS_POS_MIN INT64_MIN
+#define PRIhts_pos PRId64
+typedef int64_t hts_pos_t;
+
+// For comparison with previous release:
+//
+// #define HTS_POS_MAX INT_MAX
+// #define HTS_POS_MIN INT_MIN
+// #define PRIhts_pos PRId32
+// typedef int32_t hts_pos_t;
 
 typedef struct {
-    uint32_t beg, end;
-} hts_pair32_t;
+   hts_pos_t beg, end;
+} hts_pair_pos_t;
+
+typedef hts_pair_pos_t hts_pair32_t;  // For backwards compatibility
 
 typedef struct {
     uint64_t u, v;
@@ -527,23 +656,28 @@ typedef struct {
 
 typedef struct {
     const char *reg;
+    hts_pair_pos_t *intervals;
     int tid;
-    hts_pair32_t *intervals;
     uint32_t count;
-    uint32_t min_beg, max_end;
+    hts_pos_t min_beg, max_end;
 } hts_reglist_t;
 
-typedef int hts_readrec_func(BGZF *fp, void *data, void *r, int *tid, int *beg, int *end);
+typedef int hts_readrec_func(BGZF *fp, void *data, void *r, int *tid, hts_pos_t *beg, hts_pos_t *end);
 typedef int hts_seek_func(void *fp, int64_t offset, int where);
 typedef int64_t hts_tell_func(void *fp);
 
 typedef struct {
-    uint32_t read_rest:1, finished:1, is_cram:1, dummy:29;
-    int tid, beg, end, n_off, i;
-    int curr_tid, curr_beg, curr_end;
-    uint64_t curr_off;
-    hts_pair64_t *off;
+    uint32_t read_rest:1, finished:1, is_cram:1, nocoor:1, multi:1, dummy:27;
+    int tid, n_off, i, n_reg;
+    hts_pos_t beg, end;
+    hts_reglist_t *reg_list;
+    int curr_tid, curr_reg, curr_intv;
+    hts_pos_t curr_beg, curr_end;
+    uint64_t curr_off, nocoor_off;
+    hts_pair64_max_t *off;
     hts_readrec_func *readrec;
+    hts_seek_func *seek;
+    hts_tell_func *tell;
     struct {
         int n, m;
         int *a;
@@ -555,26 +689,74 @@ typedef struct {
     uint64_t min_off, max_off;
 } aux_key_t;
 
-typedef struct {
-    uint32_t read_rest:1, finished:1, is_cram:1, nocoor:1, dummy:28;
-    hts_reglist_t *reg_list;
-    int n_reg, i;
-    int curr_tid, curr_intv, curr_beg, curr_end, curr_reg;
-    hts_pair64_max_t *off;
-    int n_off;
-    uint64_t curr_off, nocoor_off;
-    hts_readrec_func *readrec;
-    hts_seek_func *seek;
-    hts_tell_func *tell;
-} hts_itr_multi_t;
+typedef hts_itr_t hts_itr_multi_t;
 
     #define hts_bin_first(l) (((1<<(((l)<<1) + (l))) - 1) / 7)
     #define hts_bin_parent(l) (((l) - 1) >> 3)
 
-    hts_idx_t *hts_idx_init(int n, int fmt, uint64_t offset0, int min_shift, int n_lvls);
-    void hts_idx_destroy(hts_idx_t *idx);
-    int hts_idx_push(hts_idx_t *idx, int tid, int beg, int end, uint64_t offset, int is_mapped);
-    void hts_idx_finish(hts_idx_t *idx, uint64_t final_offset);
+///////////////////////////////////////////////////////////
+// Low-level API for building indexes.
+
+/// Create a BAI/CSI/TBI type index structure
+/** @param n          Initial number of targets
+    @param fmt        Format, one of HTS_FMT_CSI, HTS_FMT_BAI or HTS_FMT_TBI
+    @param offset0    Initial file offset
+    @param min_shift  Number of bits for the minimal interval
+    @param n_lvls     Number of levels in the binning index
+    @return An initialised hts_idx_t struct on success; NULL on failure
+
+The struct returned by a successful call should be freed via hts_idx_destroy()
+when it is no longer needed.
+*/
+HTSLIB_EXPORT
+hts_idx_t *hts_idx_init(int n, int fmt, uint64_t offset0, int min_shift, int n_lvls);
+
+/// Free a BAI/CSI/TBI type index
+/** @param idx   Index structure to free
+ */
+HTSLIB_EXPORT
+void hts_idx_destroy(hts_idx_t *idx);
+
+/// Push an index entry
+/** @param idx        Index
+    @param tid        Target id
+    @param beg        Range start (zero-based)
+    @param end        Range end (zero-based, half-open)
+    @param offset     File offset
+    @param is_mapped  Range corresponds to a mapped read
+    @return 0 on success; -1 on failure
+
+The @p is_mapped parameter is used to update the n_mapped / n_unmapped counts
+stored in the meta-data bin.
+ */
+HTSLIB_EXPORT
+int hts_idx_push(hts_idx_t *idx, int tid, hts_pos_t beg, hts_pos_t end, uint64_t offset, int is_mapped);
+
+/// Finish building an index
+/** @param idx          Index
+    @param final_offset Last file offset
+    @return 0 on success; non-zero on failure.
+*/
+HTSLIB_EXPORT
+int hts_idx_finish(hts_idx_t *idx, uint64_t final_offset);
+
+/// Returns index format
+/** @param idx   Index
+    @return One of HTS_FMT_CSI, HTS_FMT_BAI or HTS_FMT_TBI
+*/
+HTSLIB_EXPORT
+int hts_idx_fmt(hts_idx_t *idx);
+
+/// Add name to TBI index meta-data
+/** @param idx   Index
+    @param tid   Target identifier
+    @param name  Target name
+    @return Index number of name in names list on success; -1 on failure.
+*/
+HTSLIB_EXPORT
+int hts_idx_tbi_name(hts_idx_t *idx, int tid, const char *name);
+
+// Index loading and saving
 
 /// Save an index to a file
 /** @param idx  Index to be written
@@ -582,6 +764,7 @@ typedef struct {
     @param fmt  One of the HTS_FMT_* index formats
     @return  0 if successful, or negative if an error occurred.
 */
+HTSLIB_EXPORT
 int hts_idx_save(const hts_idx_t *idx, const char *fn, int fmt) HTS_RESULT_USED;
 
 /// Save an index to a specific file
@@ -591,23 +774,85 @@ int hts_idx_save(const hts_idx_t *idx, const char *fn, int fmt) HTS_RESULT_USED;
     @param fmt    One of the HTS_FMT_* index formats
     @return  0 if successful, or negative if an error occurred.
 */
+HTSLIB_EXPORT
 int hts_idx_save_as(const hts_idx_t *idx, const char *fn, const char *fnidx, int fmt) HTS_RESULT_USED;
 
 /// Load an index file
 /** @param fn   BAM/BCF/etc filename, to which .bai/.csi/etc will be added or
-                the extension substituted, to search for an existing index file
+                the extension substituted, to search for an existing index file.
+                In case of a non-standard naming, the file name can include the
+                name of the index file delimited with HTS_IDX_DELIM.
     @param fmt  One of the HTS_FMT_* index formats
     @return  The index, or NULL if an error occurred.
+
+If @p fn contains the string "##idx##" (HTS_IDX_DELIM), the part before
+the delimiter will be used as the name of the data file and the part after
+it will be used as the name of the index.
+
+Otherwise, this function tries to work out the index name as follows:
+
+  It will try appending ".csi" to @p fn
+  It will try substituting an existing suffix (e.g. .bam, .vcf) with ".csi"
+  Then, if @p fmt is HTS_FMT_BAI:
+    It will try appending ".bai" to @p fn
+    To will substituting the existing suffix (e.g. .bam) with ".bai"
+  else if @p fmt is HTS_FMT_TBI:
+    It will try appending ".tbi" to @p fn
+    To will substituting the existing suffix (e.g. .vcf) with ".tbi"
+
+If the index file is remote (served over a protocol like https), first a check
+is made to see is a locally cached copy is available.  This is done for all
+of the possible names listed above.  If a cached copy is not available then
+the index will be downloaded and stored in the current working directory,
+with the same name as the remote index.
+
+    Equivalent to hts_idx_load3(fn, NULL, fmt, HTS_IDX_SAVE_REMOTE);
 */
+HTSLIB_EXPORT
 hts_idx_t *hts_idx_load(const char *fn, int fmt);
 
 /// Load a specific index file
 /** @param fn     Input BAM/BCF/etc filename
     @param fnidx  The input index filename
     @return  The index, or NULL if an error occurred.
+
+    Equivalent to hts_idx_load3(fn, fnidx, 0, 0);
+
+    This function will not attempt to save index files locally.
 */
+HTSLIB_EXPORT
 hts_idx_t *hts_idx_load2(const char *fn, const char *fnidx);
 
+/// Load a specific index file
+/** @param fn     Input BAM/BCF/etc filename
+    @param fnidx  The input index filename
+    @param fmt    One of the HTS_FMT_* index formats
+    @param flags  Flags to alter behaviour (see description)
+    @return  The index, or NULL if an error occurred.
+
+    If @p fnidx is NULL, the index name will be derived from @p fn in the
+    same way as hts_idx_load().
+
+    If @p fnidx is not NULL, @p fmt is ignored.
+
+    The @p flags parameter can be set to a combination of the following
+    values:
+
+        HTS_IDX_SAVE_REMOTE   Save a local copy of any remote indexes
+        HTS_IDX_SILENT_FAIL   Fail silently if the index is not present
+
+    The index struct returned by a successful call should be freed
+    via hts_idx_destroy() when it is no longer needed.
+*/
+HTSLIB_EXPORT
+hts_idx_t *hts_idx_load3(const char *fn, const char *fnidx, int fmt, int flags);
+
+/// Flags for hts_idx_load3() ( and also sam_idx_load3(), tbx_idx_load3() )
+#define HTS_IDX_SAVE_REMOTE 1
+#define HTS_IDX_SILENT_FAIL 2
+
+///////////////////////////////////////////////////////////
+// Functions for accessing meta-data stored in indexes
 
 /// Get extra index meta-data
 /** @param idx    The index
@@ -620,6 +865,7 @@ hts_idx_t *hts_idx_load2(const char *fn, const char *fnidx);
     the results themselves, including knowing what sort of data to expect;
     byte swapping etc.
 */
+HTSLIB_EXPORT
 uint8_t *hts_idx_get_meta(hts_idx_t *idx, uint32_t *l_meta);
 
 /// Set extra index meta-data
@@ -634,13 +880,42 @@ uint8_t *hts_idx_get_meta(hts_idx_t *idx, uint32_t *l_meta);
     If is_copy != 0, a copy of the input data is taken.  If not, ownership of
     the data pointed to by *meta passes to the index.
 */
+HTSLIB_EXPORT
 int hts_idx_set_meta(hts_idx_t *idx, uint32_t l_meta, uint8_t *meta, int is_copy);
 
-    int hts_idx_get_stat(const hts_idx_t* idx, int tid, uint64_t* mapped, uint64_t* unmapped);
-    uint64_t hts_idx_get_n_no_coor(const hts_idx_t* idx);
+/// Get number of mapped and unmapped reads from an index
+/** @param      idx      Index
+    @param      tid      Target ID
+    @param[out] mapped   Location to store number of mapped reads
+    @param[out] unmapped Location to store number of unmapped reads
+    @return 0 on success; -1 on failure (data not available)
 
+    BAI and CSI indexes store information on the number of reads for each
+    target that were mapped or unmapped (unmapped reads will generally have
+    a paired read that is mapped to the target).  This function returns this
+    infomation if it is available.
+
+    @note Cram CRAI indexes do not include this information.
+*/
+HTSLIB_EXPORT
+int hts_idx_get_stat(const hts_idx_t* idx, int tid, uint64_t* mapped, uint64_t* unmapped);
+
+/// Return the number of unplaced reads from an index
+/** @param idx    Index
+    @return Unplaced reads count
+
+    Unplaced reads are not linked to any reference (e.g. RNAME is '*' in SAM
+    files).
+*/
+HTSLIB_EXPORT
+uint64_t hts_idx_get_n_no_coor(const hts_idx_t* idx);
+
+///////////////////////////////////////////////////////////
+// Region parsing
 
 #define HTS_PARSE_THOUSANDS_SEP 1  ///< Ignore ',' separators within numbers
+#define HTS_PARSE_ONE_COORD     2  ///< chr:pos means chr:pos-pos and not chr:pos-end
+#define HTS_PARSE_LIST          4  ///< Expect a comma separated list of regions. (Disables HTS_PARSE_THOUSANDS_SEP)
 
 /// Parse a numeric string
 /** The number may be expressed in scientific notation, and optionally may
@@ -654,8 +929,25 @@ int hts_idx_set_meta(hts_idx_t *idx, uint32_t l_meta, uint8_t *meta, int is_copy
     When @a strend is NULL, a warning will be printed (if hts_verbose is HTS_LOG_WARNING
     or more) if there are any trailing characters after the number.
 */
+HTSLIB_EXPORT
 long long hts_parse_decimal(const char *str, char **strend, int flags);
 
+typedef int (*hts_name2id_f)(void*, const char*);
+typedef const char *(*hts_id2name_f)(void*, int);
+
+/// Parse a "CHR:START-END"-style region string
+/** @param str  String to be parsed
+    @param beg  Set on return to the 0-based start of the region
+    @param end  Set on return to the 1-based end of the region
+    @return  Pointer to the colon or '\0' after the reference sequence name,
+             or NULL if @a str could not be parsed.
+
+    NOTE: For compatibility with hts_parse_reg only.
+    Please use hts_parse_region instead.
+*/
+HTSLIB_EXPORT
+const char *hts_parse_reg64(const char *str, hts_pos_t *beg, hts_pos_t *end);
+
 /// Parse a "CHR:START-END"-style region string
 /** @param str  String to be parsed
     @param beg  Set on return to the 0-based start of the region
@@ -663,30 +955,209 @@ long long hts_parse_decimal(const char *str, char **strend, int flags);
     @return  Pointer to the colon or '\0' after the reference sequence name,
              or NULL if @a str could not be parsed.
 */
+HTSLIB_EXPORT
 const char *hts_parse_reg(const char *str, int *beg, int *end);
 
-    hts_itr_t *hts_itr_query(const hts_idx_t *idx, int tid, int beg, int end, hts_readrec_func *readrec);
-    void hts_itr_destroy(hts_itr_t *iter);
+/// Parse a "CHR:START-END"-style region string
+/** @param str   String to be parsed
+    @param tid   Set on return (if not NULL) to be reference index (-1 if invalid)
+    @param beg   Set on return to the 0-based start of the region
+    @param end   Set on return to the 1-based end of the region
+    @param getid Function pointer.  Called if not NULL to set tid.
+    @param hdr   Caller data passed to getid.
+    @param flags Bitwise HTS_PARSE_* flags listed above.
+    @return      Pointer to the byte after the end of the entire region
+                 specifier (including any trailing comma) on success,
+                 or NULL if @a str could not be parsed.
 
-    typedef int (*hts_name2id_f)(void*, const char*);
-    typedef const char *(*hts_id2name_f)(void*, int);
-    typedef hts_itr_t *hts_itr_query_func(const hts_idx_t *idx, int tid, int beg, int end, hts_readrec_func *readrec);
+    A variant of hts_parse_reg which is reference-id aware.  It uses
+    the iterator name2id callbacks to validate the region tokenisation works.
 
-    hts_itr_t *hts_itr_querys(const hts_idx_t *idx, const char *reg, hts_name2id_f getid, void *hdr, hts_itr_query_func *itr_query, hts_readrec_func *readrec);
-    int hts_itr_next(BGZF *fp, hts_itr_t *iter, void *r, void *data) HTS_RESULT_USED;
-    const char **hts_idx_seqnames(const hts_idx_t *idx, int *n, hts_id2name_f getid, void *hdr); // free only the array, not the values
+    This is necessary due to GRCh38 HLA additions which have reference names
+    like "HLA-DRB1*12:17".
+
+    To work around ambiguous parsing issues, eg both "chr1" and "chr1:100-200"
+    are reference names, quote using curly braces.
+    Thus "{chr1}:100-200" and "{chr1:100-200}" disambiguate the above example.
+
+    Flags are used to control how parsing works, and can be one of the below.
+
+    HTS_PARSE_THOUSANDS_SEP:
+        Ignore commas in numbers.  For example with this flag 1,234,567
+        is interpreted as 1234567.
+
+    HTS_PARSE_LIST:
+        If present, the region is assmed to be a comma separated list and
+        position parsing will not contain commas (this implicitly
+        clears HTS_PARSE_THOUSANDS_SEP in the call to hts_parse_decimal).
+        On success the return pointer will be the start of the next region, ie
+        the character after the comma.  (If *ret != '\0' then the caller can
+        assume another region is present in the list.)
+
+        If not set then positions may contain commas.  In this case the return
+        value should point to the end of the string, or NULL on failure.
+
+    HTS_PARSE_ONE_COORD:
+        If present, X:100 is treated as the single base pair region X:100-100.
+        In this case X:-100 is shorthand for X:1-100 and X:100- is X:100-<end>.
+        (This is the standard bcftools region convention.)
+
+        When not set X:100 is considered to be X:100-<end> where <end> is
+        the end of chromosome X (set to INT_MAX here).  X:100- and X:-100 are
+        invalid.
+        (This is the standard samtools region convention.)
+
+    Note the supplied string expects 1 based inclusive coordinates, but the
+    returned coordinates start from 0 and are half open, so pos0 is valid
+    for use in e.g. "for (pos0 = beg; pos0 < end; pos0++) {...}"
+
+    If NULL is returned, the value in tid mat give additional information
+    about the error:
+
+        -2   Failed to parse @p hdr; or out of memory
+        -1   The reference in @p str has mismatched braces, or does not
+             exist in @p hdr
+        >= 0 The specified range in @p str could not be parsed
+*/
+HTSLIB_EXPORT
+const char *hts_parse_region(const char *s, int *tid, hts_pos_t *beg,
+                             hts_pos_t *end, hts_name2id_f getid, void *hdr,
+                             int flags);
+
+
+///////////////////////////////////////////////////////////
+// Generic iterators
+//
+// These functions provide the low-level infrastructure for iterators.
+// Wrappers around these are used to make iterators for specific file types.
+// See:
+//     htslib/sam.h  for SAM/BAM/CRAM iterators
+//     htslib/vcf.h  for VCF/BCF iterators
+//     htslib/tbx.h  for files indexed by tabix
+
+/// Create a single-region iterator
+/** @param idx      Index
+    @param tid      Target ID
+    @param beg      Start of region
+    @param end      End of region
+    @param readrec  Callback to read a record from the input file
+    @return An iterator on success; NULL on failure
+
+    The iterator struct returned by a successful call should be freed
+    via hts_itr_destroy() when it is no longer needed.
+ */
+HTSLIB_EXPORT
+hts_itr_t *hts_itr_query(const hts_idx_t *idx, int tid, hts_pos_t beg, hts_pos_t end, hts_readrec_func *readrec);
+
+/// Free an iterator
+/** @param iter   Iterator to free
+ */
+HTSLIB_EXPORT
+void hts_itr_destroy(hts_itr_t *iter);
+
+typedef hts_itr_t *hts_itr_query_func(const hts_idx_t *idx, int tid, hts_pos_t beg, hts_pos_t end, hts_readrec_func *readrec);
+
+/// Create a single-region iterator from a text region specification
+/** @param idx       Index
+    @param reg       Region specifier
+    @param getid     Callback function to return the target ID for a name
+    @param hdr       Input file header
+    @param itr_query Callback function returning an iterator for a numeric tid,
+                     start and end position
+    @param readrec   Callback to read a record from the input file
+    @return An iterator on success; NULL on error
+
+    The iterator struct returned by a successful call should be freed
+    via hts_itr_destroy() when it is no longer needed.
+ */
+HTSLIB_EXPORT
+hts_itr_t *hts_itr_querys(const hts_idx_t *idx, const char *reg, hts_name2id_f getid, void *hdr, hts_itr_query_func *itr_query, hts_readrec_func *readrec);
+
+/// Return the next record from an iterator
+/** @param fp      Input file handle
+    @param iter    Iterator
+    @param r       Pointer to record placeholder
+    @param data    Data passed to the readrec callback
+    @return >= 0 on success, -1 when there is no more data, < -1 on error
+ */
+HTSLIB_EXPORT
+int hts_itr_next(BGZF *fp, hts_itr_t *iter, void *r, void *data) HTS_RESULT_USED;
+
+/// Return a list of target names from an index
+/** @param      idx    Index
+    @param[out] n      Location to store the number of targets
+    @param      getid  Callback function to get the name for a target ID
+    @param      hdr    Header from indexed file
+    @return An array of pointers to the names on success; NULL on failure
+
+    @note The names are pointers into the header data structure.  When cleaning
+    up, only the array should be freed, not the names.
+ */
+HTSLIB_EXPORT
+const char **hts_idx_seqnames(const hts_idx_t *idx, int *n, hts_id2name_f getid, void *hdr); // free only the array, not the values
 
 /**********************************
  * Iterator with multiple regions *
  **********************************/
 
-typedef hts_itr_multi_t *hts_itr_multi_query_func(const hts_idx_t *idx, hts_itr_multi_t *itr);
-hts_itr_multi_t *hts_itr_multi_bam(const hts_idx_t *idx, hts_itr_multi_t *iter);
-hts_itr_multi_t *hts_itr_multi_cram(const hts_idx_t *idx, hts_itr_multi_t *iter);
-hts_itr_multi_t *hts_itr_regions(const hts_idx_t *idx, hts_reglist_t *reglist, int count, hts_name2id_f getid, void *hdr, hts_itr_multi_query_func *itr_specific, hts_readrec_func *readrec, hts_seek_func *seek, hts_tell_func *tell);
-int hts_itr_multi_next(htsFile *fd, hts_itr_multi_t *iter, void *r);
+typedef int hts_itr_multi_query_func(const hts_idx_t *idx, hts_itr_t *itr);
+HTSLIB_EXPORT
+int hts_itr_multi_bam(const hts_idx_t *idx, hts_itr_t *iter);
+HTSLIB_EXPORT
+int hts_itr_multi_cram(const hts_idx_t *idx, hts_itr_t *iter);
+
+/// Create a multi-region iterator from a region list
+/** @param idx          Index
+    @param reglist      Region list
+    @param count        Number of items in region list
+    @param getid        Callback to convert names to target IDs
+    @param hdr          Indexed file header (passed to getid)
+    @param itr_specific Filetype-specific callback function
+    @param readrec      Callback to read an input file record
+    @param seek         Callback to seek in the input file
+    @param tell         Callback to return current input file location
+    @return An iterator on success; NULL on failure
+
+    The iterator struct returned by a successful call should be freed
+    via hts_itr_destroy() when it is no longer needed.
+ */
+HTSLIB_EXPORT
+hts_itr_t *hts_itr_regions(const hts_idx_t *idx, hts_reglist_t *reglist, int count, hts_name2id_f getid, void *hdr, hts_itr_multi_query_func *itr_specific, hts_readrec_func *readrec, hts_seek_func *seek, hts_tell_func *tell);
+
+/// Return the next record from an iterator
+/** @param fp      Input file handle
+    @param iter    Iterator
+    @param r       Pointer to record placeholder
+    @return >= 0 on success, -1 when there is no more data, < -1 on error
+ */
+HTSLIB_EXPORT
+int hts_itr_multi_next(htsFile *fd, hts_itr_t *iter, void *r);
+
+/// Create a region list from a char array
+/** @param argv      Char array of target:interval elements, e.g. chr1:2500-3600, chr1:5100, chr2
+    @param argc      Number of items in the array
+    @param r_count   Pointer to the number of items in the resulting region list
+    @param hdr       Header for the sam/bam/cram file
+    @param getid     Callback to convert target names to target ids.
+    @return  A region list on success, NULL on failure
+
+    The hts_reglist_t struct returned by a successful call should be freed
+    via hts_reglist_free() when it is no longer needed.
+ */
+HTSLIB_EXPORT
+hts_reglist_t *hts_reglist_create(char **argv, int argc, int *r_count, void *hdr,  hts_name2id_f getid);
+
+/// Free a region list
+/** @param reglist    Region list
+    @param count      Number of items in the list
+ */
+HTSLIB_EXPORT
 void hts_reglist_free(hts_reglist_t *reglist, int count);
-void hts_itr_multi_destroy(hts_itr_multi_t *iter);
+
+/// Free a multi-region iterator
+/** @param iter   Iterator to free
+ */
+#define hts_itr_multi_destroy(iter) hts_itr_destroy(iter)
 
 
     /**
@@ -701,6 +1172,7 @@ void hts_itr_multi_destroy(hts_itr_multi_t *iter);
     #define FT_BCF    (1<<2)
     #define FT_BCF_GZ (FT_GZ|FT_BCF)
     #define FT_STDIN  (1<<3)
+    HTSLIB_EXPORT
     int hts_file_type(const char *fname);
 
 
@@ -711,7 +1183,9 @@ void hts_itr_multi_destroy(hts_itr_multi_t *iter);
 struct errmod_t;
 typedef struct errmod_t errmod_t;
 
+HTSLIB_EXPORT
 errmod_t *errmod_init(double depcorr);
+HTSLIB_EXPORT
 void errmod_destroy(errmod_t *em);
 
 /*
@@ -720,6 +1194,7 @@ void errmod_destroy(errmod_t *em);
     bases[i]: qual:6, strand:1, base:4
     q[i*m+j]: phred-scaled likelihood of (i,j)
  */
+HTSLIB_EXPORT
 int errmod_cal(const errmod_t *em, int n, int m, uint16_t *bases, float *q);
 
 
@@ -757,6 +1232,7 @@ On failure, errno will be set to EINVAL if the values of l_ref or l_query
 were invalid; or ENOMEM if a memory allocation failed.
 */
 
+HTSLIB_EXPORT
 int probaln_glocal(const uint8_t *ref, int l_ref, const uint8_t *query, int l_query, const uint8_t *iqual, const probaln_par_t *c, int *state, uint8_t *q);
 
 
@@ -781,29 +1257,34 @@ int probaln_glocal(const uint8_t *ref, int l_ref, const uint8_t *query, int l_qu
      *
      *  @return     hts_md5_context pointer on success, NULL otherwise.
      */
+    HTSLIB_EXPORT
     hts_md5_context *hts_md5_init(void);
 
     /*! @abstract Updates the context with the MD5 of the data. */
+    HTSLIB_EXPORT
     void hts_md5_update(hts_md5_context *ctx, const void *data, unsigned long size);
 
     /*! @abstract Computes the final 128-bit MD5 hash from the given context */
+    HTSLIB_EXPORT
     void hts_md5_final(unsigned char *digest, hts_md5_context *ctx);
 
     /*! @abstract Resets an md5_context to the initial state, as returned
      *            by hts_md5_init().
      */
+    HTSLIB_EXPORT
     void hts_md5_reset(hts_md5_context *ctx);
 
     /*! @abstract Converts a 128-bit MD5 hash into a 33-byte nul-termninated
      *            hex string.
      */
+    HTSLIB_EXPORT
     void hts_md5_hex(char *hex, const unsigned char *digest);
 
     /*! @abstract Deallocates any memory allocated by hts_md5_init. */
+    HTSLIB_EXPORT
     void hts_md5_destroy(hts_md5_context *ctx);
 
-
-static inline int hts_reg2bin(int64_t beg, int64_t end, int min_shift, int n_lvls)
+static inline int hts_reg2bin(hts_pos_t beg, hts_pos_t end, int min_shift, int n_lvls)
 {
     int l, s = min_shift, t = ((1<<((n_lvls<<1) + n_lvls)) - 1) / 7;
     for (--end, l = n_lvls; l > 0; --l, s += 3, t -= 1<<((l<<1)+l))
